
#include <matrix4x4.h>

/**
 * [ a b c d ][ x ]
 * [ e f g h ][ y ]
 * [ i j k l ][ z ]
 * [ m n o p ][ w ]
 */

Vector3 multiplyVectorByMatrix( const Matrix4x4 &m, const Vector3 &v )
{
    double x, y, z;

    x =   ( m.d[ 0 ][ 0 ] * v.getX() ) + ( m.d[ 0 ][ 1 ] * v.getY() )
        + ( m.d[ 0 ][ 2 ] * v.getZ() );
    y =   ( m.d[ 1 ][ 0 ] * v.getX() ) + ( m.d[ 1 ][ 1 ] * v.getY() )
        + ( m.d[ 1 ][ 2 ] * v.getZ() );
    z =   ( m.d[ 2 ][ 0 ] * v.getX() ) + ( m.d[ 2 ][ 1 ] * v.getY() )
        + ( m.d[ 2 ][ 2 ] * v.getZ() );

    return Vector3( x, y, z );
}

Point3 multiplyPointByMatrix( const Matrix4x4 &m, const Point3 &v )
{
    double x, y, z;

    x =   ( m.d[ 0 ][ 0 ] * v.getX() ) + ( m.d[ 0 ][ 1 ] * v.getY() )
        + ( m.d[ 0 ][ 2 ] * v.getZ() ) + ( m.d[ 0 ][ 3 ] );
    y =   ( m.d[ 1 ][ 0 ] * v.getX() ) + ( m.d[ 1 ][ 1 ] * v.getY() )
        + ( m.d[ 1 ][ 2 ] * v.getZ() ) + ( m.d[ 1 ][ 3 ] );
    z =   ( m.d[ 2 ][ 0 ] * v.getX() ) + ( m.d[ 2 ][ 1 ] * v.getY() )
        + ( m.d[ 2 ][ 2 ] * v.getZ() ) + ( m.d[ 2 ][ 3 ] );

    return Point3( x, y, z );
}

Matrix4x4 multiplyMatrix( const Matrix4x4 &m1, const Matrix4x4 &m2 )
{
    Matrix4x4 mr;
    int i, j;

    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 4; j++ ) {
            mr.d[ i ][ j ] =   (m1.d[i][0] * m2.d[0][j])
                             + (m1.d[i][1] * m2.d[1][j])
                             + (m1.d[i][2] * m2.d[2][j])
                             + (m1.d[i][3] * m2.d[3][j]);
        }
    }

    return mr;
}

Matrix4x4 buildObjectToWorld( const Point3 &position,
                              const Vector3 &forward,
                              const Vector3 &up )
{
    Matrix4x4 r;
    Vector3 right = forward.cross( up );

    r.d[0][0] = right.getX();
    r.d[0][1] = right.getY();
    r.d[0][2] = right.getZ();
    r.d[0][3] = position.getX();

    r.d[1][0] = up.getX();
    r.d[1][1] = up.getY();
    r.d[1][2] = up.getZ();
    r.d[1][3] = position.getY();

    r.d[2][0] = forward.getX();
    r.d[2][1] = forward.getY();
    r.d[2][2] = forward.getZ();
    r.d[2][3] = position.getZ();

    r.d[3][0] = 0.0;
    r.d[3][1] = 0.0;
    r.d[3][2] = 0.0;
    r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 buildTranslation( const double &x, const double &y, const double &z )
{
    Matrix4x4 r;

    r.d[0][0] = 1.0; r.d[0][1] = 0.0; r.d[0][2] = 0.0; r.d[0][3] = x;
    r.d[1][0] = 0.0; r.d[1][1] = 1.0; r.d[1][2] = 0.0; r.d[1][3] = y;
    r.d[2][0] = 0.0; r.d[2][1] = 0.0; r.d[2][2] = 1.0; r.d[2][3] = z;
    r.d[3][0] = 0.0; r.d[3][1] = 0.0; r.d[3][2] = 0.0; r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 buildScale( const double &sx, const double &sy, const double &sz )
{
    Matrix4x4 r;

    r.d[0][0] = sx;  r.d[0][1] = 0.0; r.d[0][2] = 0.0; r.d[0][3] = 0.0;
    r.d[1][0] = 0.0; r.d[1][1] = sy;  r.d[1][2] = 0.0; r.d[1][3] = 0.0;
    r.d[2][0] = 0.0; r.d[2][1] = 0.0; r.d[2][2] = sz;  r.d[2][3] = 0.0;
    r.d[3][0] = 0.0; r.d[3][1] = 0.0; r.d[3][2] = 0.0; r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 identityMatrix( void )
{
    Matrix4x4 r;

    r.d[0][0] = 1.0; r.d[0][1] = 0.0; r.d[0][2] = 0.0; r.d[0][3] = 0.0;
    r.d[1][0] = 0.0; r.d[1][1] = 1.0; r.d[1][2] = 0.0; r.d[1][3] = 0.0;
    r.d[2][0] = 0.0; r.d[2][1] = 0.0; r.d[2][2] = 1.0; r.d[2][3] = 0.0;
    r.d[3][0] = 0.0; r.d[3][1] = 0.0; r.d[3][2] = 0.0; r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 buildRotationX( const double &angle )
{
    Matrix4x4 r;

    r.d[0][0] = 1.0; r.d[0][1] = 0.0;          r.d[0][2] = 0.0;           r.d[0][3] = 0.0;
    r.d[1][0] = 0.0; r.d[1][1] = cos( angle ); r.d[1][2] = -sin( angle ); r.d[1][3] = 0.0;
    r.d[2][0] = 0.0; r.d[2][1] = sin( angle ); r.d[2][2] = cos( angle );  r.d[2][3] = 0.0;
    r.d[3][0] = 0.0; r.d[3][1] = 0.0;          r.d[3][2] = 0.0;           r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 buildRotationY( const double &angle )
{
    Matrix4x4 r;

    r.d[0][0] = cos( angle );  r.d[0][1] = 0.0; r.d[0][2] = sin( angle ); r.d[0][3] = 0.0;
    r.d[1][0] = 0.0;           r.d[1][1] = 1.0; r.d[1][2] = 0.0;          r.d[1][3] = 0.0;
    r.d[2][0] = -sin( angle ); r.d[2][1] = 0.0; r.d[2][2] = cos( angle ); r.d[2][3] = 0.0;
    r.d[3][0] = 0.0;           r.d[3][1] = 0.0; r.d[3][2] = 0.0;          r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 buildRotationZ( const double &angle )
{
    Matrix4x4 r;

    r.d[0][0] = cos( angle ); r.d[0][1] = -sin( angle ); r.d[0][2] = 0.0; r.d[0][3] = 0.0;
    r.d[1][0] = sin( angle ); r.d[1][1] = cos( angle );  r.d[1][2] = 0.0; r.d[1][3] = 0.0;
    r.d[2][0] = 0.0;          r.d[2][1] = 0.0;           r.d[2][2] = 1.0; r.d[2][3] = 0.0;
    r.d[3][0] = 0.0;          r.d[3][1] = 0.0;           r.d[3][2] = 0.0; r.d[3][3] = 1.0;

    return r;
}

Matrix4x4 transposeMatrix( const Matrix4x4 &m )
{
    Matrix4x4 r;
    int i, j;

    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 4; j++ ) {
            r.d[ i ][ j ] = m.d[ j ][ i ];
        }
    }

    return r;
}

/**
 * The code for this is from the Intel tech note AP-928 on using SIMD for
 * finding the inverse of 4x4 matricies.  I'll do the SIMD version sometime but
 * to be honest I doubt I'll be calling this alot except in precalculations.
 */
Matrix4x4 inverseMatrix( const Matrix4x4 &m )
{
    Matrix4x4 r, st;
    double tmp[ 12 ];
    double det;
    int i, j;

    st = transposeMatrix( m );

    /* Calculate pairs for the first 8 elements (cofactors) */
    tmp[  0 ] = st.d[ 2 ][ 2 ] * st.d[ 3 ][ 3 ];
    tmp[  1 ] = st.d[ 2 ][ 3 ] * st.d[ 3 ][ 2 ];
    tmp[  2 ] = st.d[ 2 ][ 1 ] * st.d[ 3 ][ 3 ];
    tmp[  3 ] = st.d[ 2 ][ 3 ] * st.d[ 3 ][ 1 ];
    tmp[  4 ] = st.d[ 2 ][ 1 ] * st.d[ 3 ][ 2 ];
    tmp[  5 ] = st.d[ 2 ][ 2 ] * st.d[ 3 ][ 1 ];
    tmp[  6 ] = st.d[ 2 ][ 0 ] * st.d[ 3 ][ 3 ];
    tmp[  7 ] = st.d[ 2 ][ 3 ] * st.d[ 3 ][ 0 ];
    tmp[  8 ] = st.d[ 2 ][ 0 ] * st.d[ 3 ][ 2 ];
    tmp[  9 ] = st.d[ 2 ][ 2 ] * st.d[ 3 ][ 0 ];
    tmp[ 10 ] = st.d[ 2 ][ 0 ] * st.d[ 3 ][ 1 ];
    tmp[ 11 ] = st.d[ 2 ][ 1 ] * st.d[ 3 ][ 0 ];

    /* calculate first 8 elements (cofactors) */
    r.d[0][0] =   (tmp[0] * st.d[1][1]) + (tmp[3] * st.d[1][2]) + (tmp[ 4] * st.d[1][3])
                - (tmp[1] * st.d[1][1]) - (tmp[2] * st.d[1][2]) - (tmp[ 5] * st.d[1][3]);
    r.d[0][1] =   (tmp[1] * st.d[1][1]) + (tmp[6] * st.d[1][2]) + (tmp[ 9] * st.d[1][3])
                - (tmp[0] * st.d[1][0]) - (tmp[7] * st.d[1][2]) - (tmp[ 8] * st.d[1][3]);
    r.d[0][2] =   (tmp[2] * st.d[1][0]) + (tmp[7] * st.d[1][1]) + (tmp[10] * st.d[1][3])
                - (tmp[3] * st.d[1][0]) - (tmp[6] * st.d[1][1]) - (tmp[11] * st.d[1][3]);
    r.d[0][3] =   (tmp[5] * st.d[1][0]) + (tmp[8] * st.d[1][1]) + (tmp[11] * st.d[1][2])
                - (tmp[4] * st.d[1][0]) - (tmp[9] * st.d[1][1]) - (tmp[10] * st.d[1][2]);
    r.d[1][0] =   (tmp[1] * st.d[0][1]) + (tmp[2] * st.d[0][2]) + (tmp[ 5] * st.d[0][3])
                - (tmp[0] * st.d[0][1]) - (tmp[3] * st.d[0][2]) - (tmp[ 4] * st.d[0][3]);
    r.d[1][1] =   (tmp[0] * st.d[0][0]) + (tmp[7] * st.d[0][2]) + (tmp[ 8] * st.d[0][3])
                - (tmp[1] * st.d[0][0]) - (tmp[6] * st.d[0][2]) - (tmp[ 9] * st.d[0][3]);
    r.d[1][2] =   (tmp[3] * st.d[0][0]) + (tmp[6] * st.d[0][1]) + (tmp[11] * st.d[0][3])
                - (tmp[2] * st.d[0][0]) - (tmp[7] * st.d[0][1]) - (tmp[10] * st.d[0][3]);
    r.d[1][3] =   (tmp[4] * st.d[0][0]) + (tmp[9] * st.d[0][1]) + (tmp[10] * st.d[0][2])
                - (tmp[5] * st.d[0][0]) - (tmp[8] * st.d[0][1]) - (tmp[11] * st.d[0][2]);

    /* Calculate pairs for the second 8 elements (cofactors) */
    tmp[  0 ] = st.d[ 0 ][ 2 ] * st.d[ 1 ][ 3 ];
    tmp[  1 ] = st.d[ 0 ][ 3 ] * st.d[ 1 ][ 2 ];
    tmp[  2 ] = st.d[ 0 ][ 1 ] * st.d[ 1 ][ 3 ];
    tmp[  3 ] = st.d[ 0 ][ 3 ] * st.d[ 1 ][ 1 ];
    tmp[  4 ] = st.d[ 0 ][ 1 ] * st.d[ 1 ][ 2 ];
    tmp[  5 ] = st.d[ 0 ][ 2 ] * st.d[ 1 ][ 1 ];
    tmp[  6 ] = st.d[ 0 ][ 0 ] * st.d[ 1 ][ 3 ];
    tmp[  7 ] = st.d[ 0 ][ 3 ] * st.d[ 1 ][ 0 ];
    tmp[  8 ] = st.d[ 0 ][ 0 ] * st.d[ 1 ][ 2 ];
    tmp[  9 ] = st.d[ 0 ][ 2 ] * st.d[ 1 ][ 0 ];
    tmp[ 10 ] = st.d[ 0 ][ 0 ] * st.d[ 1 ][ 1 ];
    tmp[ 11 ] = st.d[ 0 ][ 1 ] * st.d[ 1 ][ 0 ];

    /* calculate second 8 elements (cofactors) */
    r.d[2][0] =   (tmp[ 0] * st.d[3][1]) + (tmp[ 3] * st.d[3][2]) + (tmp[ 4] * st.d[3][3])
                - (tmp[ 1] * st.d[3][1]) - (tmp[ 2] * st.d[3][2]) - (tmp[ 5] * st.d[3][3]);
    r.d[2][1] =   (tmp[ 1] * st.d[3][0]) + (tmp[ 6] * st.d[3][2]) + (tmp[ 9] * st.d[3][3])
                - (tmp[ 0] * st.d[3][0]) - (tmp[ 7] * st.d[3][2]) - (tmp[ 8] * st.d[3][3]);
    r.d[2][2] =   (tmp[ 2] * st.d[3][0]) + (tmp[ 7] * st.d[3][1]) + (tmp[10] * st.d[3][3])
                - (tmp[ 3] * st.d[3][0]) - (tmp[ 6] * st.d[3][1]) - (tmp[11] * st.d[3][3]);
    r.d[2][3] =   (tmp[ 5] * st.d[3][0]) + (tmp[ 8] * st.d[3][1]) + (tmp[11] * st.d[3][2])
                - (tmp[ 4] * st.d[3][0]) - (tmp[ 9] * st.d[3][1]) - (tmp[10] * st.d[3][2]);
    r.d[3][0] =   (tmp[ 2] * st.d[2][2]) + (tmp[ 5] * st.d[2][3]) + (tmp[ 1] * st.d[2][1])
                - (tmp[ 4] * st.d[2][3]) - (tmp[ 0] * st.d[2][1]) - (tmp[ 3] * st.d[2][2]);
    r.d[3][1] =   (tmp[ 8] * st.d[2][3]) + (tmp[ 0] * st.d[2][0]) + (tmp[ 7] * st.d[2][2])
                - (tmp[ 6] * st.d[2][2]) - (tmp[ 9] * st.d[2][3]) - (tmp[ 1] * st.d[2][0]);
    r.d[3][2] =   (tmp[ 6] * st.d[2][1]) + (tmp[11] * st.d[2][3]) + (tmp[ 3] * st.d[2][0])
                - (tmp[10] * st.d[2][3]) - (tmp[ 2] * st.d[2][0]) - (tmp[ 7] * st.d[2][1]);
    r.d[3][3] =   (tmp[10] * st.d[2][2]) + (tmp[ 4] * st.d[2][0]) + (tmp[ 9] * st.d[2][1])
                - (tmp[ 8] * st.d[2][1]) - (tmp[11] * st.d[2][2]) - (tmp[ 5] * st.d[2][0]);

    /* Calculate determinant. */
    det =   (st.d[ 0 ][ 0 ] * r.d[ 0 ][ 0 ])
          + (st.d[ 0 ][ 1 ] * r.d[ 0 ][ 1 ])
          + (st.d[ 0 ][ 2 ] * r.d[ 0 ][ 2 ])
          + (st.d[ 0 ][ 3 ] * r.d[ 0 ][ 3 ]);

    /* Finalize inverse. */
    det = 1.0 / det;
    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 4; j++ ) {
            r.d[ i ][ j ] *= det;
        }
    }

    return r;
}

