
#include <math.h>
#include <stdio.h>
#include <primitive.h>
#include <pointlight.h>
#include <scene.h>

#define USE_GRID

Scene::Scene( BaseObject *scene, int gridwidth )
    : grid( gridwidth ), gridw( gridwidth )
{
    gridscale = 1.0;
    addScene( scene );
}

Scene::~Scene( void )
{
}

void Scene::addScene( BaseObject *root )
{
    BoundingBox b;
    PrimitiveList plist = root->foldModel( identityMatrix() );
    PrimitiveList::const_iterator it;

    b.minx = HUGE_VAL; b.maxx = -HUGE_VAL;
    b.miny = HUGE_VAL; b.maxy = -HUGE_VAL;
    b.minz = HUGE_VAL; b.maxz = -HUGE_VAL;

    for( it = plist.begin(); it != plist.end(); it++ ) {
        if( (*it)->isPointLight() ) continue;

        BoundingBox curbox = (*it)->getBoundingBox();
        if( b.minx > curbox.minx ) b.minx = curbox.minx;
        if( b.maxx < curbox.maxx ) b.maxx = curbox.maxx;
        if( b.miny > curbox.miny ) b.miny = curbox.miny;
        if( b.maxy < curbox.maxy ) b.maxy = curbox.maxy;
        if( b.minz > curbox.minz ) b.minz = curbox.minz;
        if( b.maxz < curbox.maxz ) b.maxz = curbox.maxz;
    }
    double xlength = b.maxx - b.minx;
    double ylength = b.maxy - b.miny;
    double zlength = b.maxz - b.minz;
    double bestlength = xlength;
    if( ylength > bestlength ) bestlength = ylength;
    if( zlength > bestlength ) bestlength = zlength;
    gridscale = ( (double) gridw ) / bestlength;

    fprintf( stderr, "scene: Scaling scene by factor %f "
                     "for better partitioning.\n", gridscale );
    plist = root->foldModel( buildScale( gridscale, gridscale, gridscale ) );

    for( it = plist.begin(); it != plist.end(); it++ ) {
        if( !(*it)->getMaterial() ) {
            fprintf( stderr, "scene: Error!  No material specified "
                             "for some object!\n" );
            continue;
        }
        if( (*it)->getMaterial()->isLuminous() ) {
            lights.push_back( (*it) );
        }
        objects.push_back( (*it) );
        if( !((*it)->isPointLight()) ) grid.addPrimitive( (*it) );
    }
    grid.printBounds();
}

void Scene::setCamera( Camera *newc )
{
    c = newc;
    c->transformEyePosition( buildScale( gridscale, gridscale, gridscale ) );
    grid.addPrimitive( new PointLight( c->getEye() ) );
}

#ifdef USE_GRID
bool Scene::closestIntersect( const Point3 &origin, const Vector3 &dir,
                              double *t, Vector3 *norm, Material **mat ) const
{
    Point3 curpoint = origin; // .add( dir.scalarMult( epsilon ) );

    while( !grid.gonFishin( curpoint ) ) {
        GridSquare *sq = grid.getCurrentSquare( curpoint );
        if( sq ) {
            PrimitiveList::const_iterator it;
            bool intersect = false;
            Vector3 normal;
            Material *material;
            double curt;

            for( it = sq->contents.begin(); it != sq->contents.end(); it++ ) {
                if( (*it)->closestIntersect( origin, dir, &curt,
                                             &normal, &material ) ) {
                    if( !intersect || curt < *t ) {
                        if( grid.getCurrentSquare( origin.add( dir.scalarMult( curt + epsilon ) ) ) == sq ) {
                            *t = curt;
                            *norm = normal;
                            *mat = material;
                            intersect = true;
                        }
                    }
                }
            }
            if( intersect ) return true;
        }

        // Nothing in the current square, move along.
        double newt = grid.getNextSquare( curpoint, dir );
        curpoint = curpoint.add( dir.scalarMult( newt ) );
    }

    return false;
}
#else
bool Scene::closestIntersect( const Point3 &origin, const Vector3 &dir,
                              double *t, Vector3 *norm, Material **mat ) const
{
    PrimitiveList::const_iterator it;
    bool intersect = false;
    Vector3 normal;
    Material *material;
    double curt;

    for( it = objects.begin(); it != objects.end(); it++ ) {
        if( (*it)->closestIntersect( origin, dir, &curt, &normal, &material ) ) {
            if( !intersect || curt < *t ) {
                *t = curt;
                *norm = normal;
                *mat = material;
                intersect = true;
            }
        }
    }

    return intersect;
}
#endif

