
#include <stl.h>
#include <string>
#include <string.h>
#include <stdio.h>
#include <node.h>
#include <vertexlist.h>
#include <primitive.h>
#include <sphere.h>
#include <convexpoly.h>
#include <triangle.h>
#include <pointlight.h>
#include <sceneparser.h>

using namespace std;

SceneParser::SceneParser( const char *filename )
    : width( 640 ), height( 480 ), gridwidth( 10 )
{
    string curtoken;

    infile.open( filename );
    if( infile.bad() ) {
        fprintf( stderr, "sceneparser: Can't open file.\n" );
        return;
    }

    while( !infile.eof() ) {
        infile >> curtoken;
        if( infile.eof() ) break;
        if( curtoken == "material" ) {
            parseMaterial();
        } else if( curtoken == "node" ) {
            nodes.push_back( parseNode() );
        } else if( curtoken == "camera" ) {
            parseCamera();
        } else if( curtoken == "width" ) {
            infile >> curtoken;
            width = atoi( curtoken.c_str() );
        } else if( curtoken == "gridwidth" ) {
            infile >> curtoken;
            gridwidth = atoi( curtoken.c_str() );
        } else if( curtoken == "height" ) {
            infile >> curtoken;
            height = atoi( curtoken.c_str() );
        }
    }
}

SceneParser::~SceneParser( void )
{
}

void SceneParser::readTriple( double &x, double &y, double &z )
{
    string curtoken;

    infile >> curtoken;
    x = atof( curtoken.c_str() );
    infile >> curtoken;
    y = atof( curtoken.c_str() );
    infile >> curtoken;
    z = atof( curtoken.c_str() );
}

void SceneParser::parseMaterial( void )
{
    Material *m = new Material;
    string curtoken;
    string matname;

    infile >> matname;
    m->setName( matname.c_str() );

    while( !infile.eof() ) {
        infile >> curtoken;
        if( curtoken == "endmaterial" ) break;
        if( curtoken == "rgb" ) {
            double r, g, b;
            readTriple( r, g, b );
            m->setColour( PixelXYZ::createPixelRGB( r, g, b ) );
        } else if( curtoken == "xyz" ) {
            double x, y, z;
            readTriple( x, y, z );
            m->setColour( PixelXYZ( x, y, z ) );
        } else if( curtoken == "reflectivity" ) {
            infile >> curtoken;
            m->setReflectivity( atof( curtoken.c_str() ) );
        } else if( curtoken == "luminous" ) {
            m->setLuminous( true );
        }
    }
    materials.push_back( m );
}

void SceneParser::parseCamera( void )
{
    Point3 pos( 0.0, 0.0, 40.0 );
    Vector3 up( 0.0, 1.0, 0.0 );
    Vector3 fwd( 0.0, 0.0, -1.0 );
    double fov = 30.0;
    string curtoken;

    while( !infile.eof() ) {
        infile >> curtoken;
        if( curtoken == "endcamera" ) break;
        if( curtoken == "up" ) {
            double x, y, z;
            readTriple( x, y, z );
            up = Vector3( x, y, z );
        } else if( curtoken == "forward" ) {
            double x, y, z;
            readTriple( x, y, z );
            fwd = Vector3( x, y, z );
        } else if( curtoken == "eyepos" ) {
            double x, y, z;
            readTriple( x, y, z );
            pos = Point3( x, y, z );
        } else if( curtoken == "fov" ) {
            infile >> curtoken;
            fov = atof( curtoken.c_str() );
        }
    }

    camera = new Camera( pos, up, fwd, fov );
}

Material *SceneParser::getMaterial( const char *name ) const
{
    vector<Material *>::const_iterator it;

    for( it = materials.begin(); it != materials.end(); it++ ) {
        if( strcmp( name, (*it)->getName() ) == 0 ) {
            return (*it);
        }
    }

    return 0;
}

Node *SceneParser::getNode( const char *name ) const
{
    vector<Node *>::const_iterator it;

    for( it = nodes.begin(); it != nodes.end(); it++ ) {
        if( strcmp( name, (*it)->getName() ) == 0 ) {
            return (*it);
        }
    }

    return 0;
}

Node *SceneParser::parseNode( void )
{
    VertexList vlist;
    Material *m = 0;
    bool moverride = false;
    Node *n = new Node;
    string curtoken;
    string nodename;

    infile >> nodename;
    n->setName( nodename.c_str() );

    while( !infile.eof() ) {
        infile >> curtoken;
        if( curtoken == "endnode" ) break;
        if( curtoken == "translate" ) {
            double x, y, z;
            readTriple( x, y, z );
            n->applyMatrix( buildTranslation( x, y, z ) );
        } else if( curtoken == "scale" ) {
            double x, y, z;
            readTriple( x, y, z );
            n->applyMatrix( buildScale( x, y, z ) );
        } else if( curtoken == "rotatex" ) {
            double a;
            infile >> curtoken;
            a = atof( curtoken.c_str() );
            n->applyMatrix( buildRotationX( a ) );
        } else if( curtoken == "rotatey" ) {
            double a;
            infile >> curtoken;
            a = atof( curtoken.c_str() );
            n->applyMatrix( buildRotationY( a ) );
        } else if( curtoken == "rotatez" ) {
            double a;
            infile >> curtoken;
            a = atof( curtoken.c_str() );
            n->applyMatrix( buildRotationZ( a ) );
        } else if( curtoken == "surface" ) {
            infile >> curtoken;
            m = getMaterial( curtoken.c_str() );
            if( !m ) {
                fprintf( stderr, "sceneparser: Material '%s' not found.\n",
                         curtoken.c_str() );
            }
        } else if( curtoken == "surfaceoverride" ) {
            moverride = true;
        } else if( curtoken == "node" ) {
            n->addObject( parseNode() );
        } else if( curtoken == "pointlight" ) {
            n->addObject( new PointLight );
        } else if( curtoken == "instance" ) {
            infile >> curtoken;
            Node *in = getNode( curtoken.c_str() );
            if( in ) {
                n->addObject( in );
            } else {
                fprintf( stderr, "sceneparser: Instance '%s' not found.\n",
                                 curtoken.c_str() );
            }
        } else if( curtoken == "sphere" ) {
            n->addObject( new Sphere );
        } else if( curtoken == "verticies" ) {
            infile >> curtoken; // read {
            while( !infile.eof() ) {
                double x, y, z;
                infile >> curtoken;
                if( curtoken == "}" ) break;
                readTriple( x, y, z );
                vlist.push_back( Point3( x, y, z ) );
                infile >> curtoken; // read }
            }
        } else if( curtoken == "polygons" ) {
            infile >> curtoken; // read {
            while( !infile.eof() ) {
                VertexList plist;
                infile >> curtoken;
                if( curtoken == "}" ) break;

                // Read until end of polygon.
                while( !infile.eof() ) {
                    int curval;
                    infile >> curtoken;
                    if( curtoken == "}" ) break;
                    curval = atoi( curtoken.c_str() ) - 1;
                    plist.push_back( vlist[ curval ] );
                }

                // Got a polygon, make either a convex poly or triangle.
                if( plist.size() == 3 ) {
                    n->addObject( new Triangle( &plist ) );
                } else {
                    n->addObject( new ConvexPoly( &plist ) );
                }
            }
        }
    }
    if( m ) n->setMaterial( m, moverride );

    return n;
}

