
#include <assert.h>
#include <stdio.h>
#include <primitive.h>
#include <shader.h>

Shader::Shader( const Scene *scene )
    : s( scene )
{
}

Shader::~Shader( void )
{
}

PixelXYZ Shader::shadeRay( const Point3 &origin, const Vector3 &dir ) const
{
    Material *mat;
    Vector3 norm;
    double t;

    if( !s->closestIntersect( origin, dir, &t, &norm, &mat ) ) {
        return PixelXYZ::black;
    }

    return shadePoint( origin, origin.add( dir.scalarMult( t ) ),
                       norm, mat );
}

PixelXYZ Shader::shadePoint( const Point3 &viewer,
                             const Point3 &pos, const Vector3 &norm,
                             Material *mat ) const
{
    PixelXYZ result = PixelXYZ::black;
    PrimitiveList::const_iterator it;

    for( it = s->getLights().begin(); it != s->getLights().end(); it++ ) {
        Point3 lpos = (*it)->getRandomLightPosition( pos );
        Vector3 tolight = lpos.subtract( pos );
        Point3 offpoint = pos.add( tolight.scalarMult( 5.0 * epsilon ) );
        Vector3 curnorm;
        Material *curmat;
        double t;

        if( s->closestIntersect( offpoint, tolight, &t, &curnorm, &curmat ) ) {
            if( t <= 1.0 ) continue;
        }

        Vector3 normnorm = norm.normalize();
        Vector3 lightnorm = tolight.normalize();

        /* alright, phriggin phong time */
        Vector3 reflection = normnorm.scalarMult( 2.0 * normnorm.dotProduct( lightnorm ) );
        Vector3 toeye = viewer.subtract( pos ).normalize();
        reflection = reflection.subtract( lightnorm ).normalize();
        double dot = reflection.dotProduct( toeye );
        if( dot < 0.0 ) dot = 0.0;
        dot = pow( dot, 50.0 );
        if( isinf( dot ) || isnan( dot ) ) dot = 0.0;
        PixelXYZ pcolour = (*it)->getMaterial()->getColour().scalarMult( dot );
        result = result.add( pcolour );

        /* and lame lamb */
        double ldot = normnorm.dotProduct( lightnorm );
        if( ldot < 0.0 ) ldot = 0.0;
        if( isinf( ldot ) || isnan( ldot ) ) ldot = 0.0;
        PixelXYZ lcolour = mat->getColour().scalarMult( ldot ); // / tolight.length() );
        result = result.add( lcolour );
    }

    return result;
}

PixelXYZ Shader::pointSample( double x, double y ) const
{
    /* Our output pixel starts at black. */
    PixelXYZ result = PixelXYZ::black;

    /* Trace a ray! XXX */
    Point3 curpos = s->getCamera()->getEye();
    Vector3 curdir = s->getCamera()->screenRay( x, y );

    result = shadeRay( curpos, curdir );

    return result;
}

PixelXYZ Shader::adaptivePoint( double x, double y, int depth,
                                double w, double h, int maxdepth ) const
{
    if( depth > maxdepth ) return pointSample( x / w, y / h );

    PixelXYZ result = PixelXYZ::black;

    double left  = x - ( 1.0 / ( 2 * depth ) );
    double right = x + ( 1.0 / ( 2 * depth ) );
    double top   = y + ( 1.0 / ( 2 * depth ) );
    double bot   = y - ( 1.0 / ( 2 * depth ) );

    PixelXYZ topleft  = pointSample(  left / w, top / h );
    PixelXYZ topright = pointSample( right / w, top / h );
    PixelXYZ botleft  = pointSample(  left / w, bot / h );
    PixelXYZ botright = pointSample( right / w, bot / h );

    PixelXYZ average = topleft.add( topright.add( botleft.add( botright ) ) );
    average = average.scalarMult( 0.25 );

    if(  topleft.subtract( average ).intensitySquared() > 0.002 ||
        topright.subtract( average ).intensitySquared() > 0.002 ||
         botleft.subtract( average ).intensitySquared() > 0.002 ||
        botright.subtract( average ).intensitySquared() > 0.002 ) {

        result = result.add( adaptivePoint( left + ( ( x - left ) / 2.0 ),
                                            y + ( ( top - y ) / 2.0 ),
                                            depth + 1, w, h, maxdepth ) );
        result = result.add( adaptivePoint( x + ( ( right - x ) / 2.0 ),
                                            y + ( ( top - y ) / 2.0 ),
                                            depth + 1, w, h, maxdepth ) );
        result = result.add( adaptivePoint( left + ( ( x - left ) / 2.0 ),
                                            y - ( ( top - y ) / 2.0 ),
                                            depth + 1, w, h, maxdepth ) );
        result = result.add( adaptivePoint( x + ( ( right - x ) / 2.0 ),
                                            y - ( ( top - y ) / 2.0 ),
                                            depth + 1, w, h, maxdepth ) );
        result = result.scalarMult( 0.25 );
    } else {
        result = average;
    }

    return result;
}

PixelXYZ Shader::superPoint( double x, double y, int depth,
                          double w, double h, int maxdepth ) const
{
    if( depth > maxdepth ) return pointSample( x / w, y / h );

    PixelXYZ result = PixelXYZ::black;
    double left  = x - ( 1.0 / ( 2 * depth ) );
    double right = x + ( 1.0 / ( 2 * depth ) );
    double top   = y + ( 1.0 / ( 2 * depth ) );

    result = result.add( superPoint( left + ( ( x - left ) / 2.0 ),
                                     y + ( ( top - y ) / 2.0 ),
                                     depth + 1, w, h, maxdepth ) );
    result = result.add( superPoint( x + ( ( right - x ) / 2.0 ),
                                     y + ( ( top - y ) / 2.0 ),
                                     depth + 1, w, h, maxdepth ) );
    result = result.add( superPoint( left + ( ( x - left ) / 2.0 ),
                                     y - ( ( top - y ) / 2.0 ),
                                     depth + 1, w, h, maxdepth ) );
    result = result.add( superPoint( x + ( ( right - x ) / 2.0 ),
                                     y - ( ( top - y ) / 2.0 ),
                                     depth + 1, w, h, maxdepth ) );
    result = result.scalarMult( 0.25 );

    return result;
}

void Shader::adaptiveSampleScanline( PixelXYZ *output, int y,
                                     int w, int h, int depth ) const
{
    int x;

    for( x = 0; x < w; x++ ) {
        *output++ = adaptivePoint( x, y, 1, w, h, depth );
    }
}

void Shader::superSampleScanline( PixelXYZ *output, int y,
                                  int w, int h, int depth ) const
{
    int x;

    for( x = 0; x < w; x++ ) {
        *output++ = superPoint( x, y, 1, w, h, depth );
    }
}

void Shader::pointSampleScanline( PixelXYZ *output, int y, int w, int h ) const
{
    int x;

    for( x = 0; x < w; x++ ) {
        *output++ = pointSample( (double) x / (double) w,
                                 (double) y / (double) h );
    }
}

