
/**
 * Triangle intersection routine from:
 *
 * Fast, Minimum Storage Ray-Triangle Intersection
 * Tomas Mller, Prosolvia Clarus AB, Sweden, tompa@clarus.se
 * Ben Trumbore, Cornell University, Ithaca, New York, wbt@graphics.cornell.edu
 *
 * http://www.acm.org/jgt/papers/MollerTrumbore97/
 */

#include <utils.h>
#include <triangle.h>

Triangle::Triangle( VertexList *vlist )
{
    ptlist[ 0 ] = (*vlist)[ 0 ];
    ptlist[ 1 ] = (*vlist)[ 1 ];
    ptlist[ 2 ] = (*vlist)[ 2 ];

    edge1 = ptlist[ 1 ].subtract( ptlist[ 0 ] );
    edge2 = ptlist[ 2 ].subtract( ptlist[ 0 ] );

    Vector3 v1 = edge1.normalize();
    Vector3 v2 = edge2.normalize();
    normal = v1.cross( v2 ).normalize();
}

Triangle::~Triangle( void )
{
}

BoundingBox Triangle::getBoundingBox( void ) const
{
    BoundingBox b;
    int i;

    b.minx = HUGE_VAL; b.maxx = -HUGE_VAL;
    b.miny = HUGE_VAL; b.maxy = -HUGE_VAL;
    b.minz = HUGE_VAL; b.maxz = -HUGE_VAL;

    for( i = 0; i < 3; i++ ) {
        if( ptlist[ i ].getX() < b.minx ) b.minx = ptlist[ i ].getX();
        if( ptlist[ i ].getX() > b.maxx ) b.maxx = ptlist[ i ].getX();

        if( ptlist[ i ].getY() < b.miny ) b.miny = ptlist[ i ].getY();
        if( ptlist[ i ].getY() > b.maxy ) b.maxy = ptlist[ i ].getY();

        if( ptlist[ i ].getZ() < b.minz ) b.minz = ptlist[ i ].getZ();
        if( ptlist[ i ].getZ() > b.maxz ) b.maxz = ptlist[ i ].getZ();
    }

    return b;
}

PrimitiveList Triangle::foldModel( const Matrix4x4 &xform ) const
{
    PrimitiveList pl;
    VertexList newlist;

    newlist.push_back( multiplyPointByMatrix( xform, ptlist[ 0 ] ) );
    newlist.push_back( multiplyPointByMatrix( xform, ptlist[ 1 ] ) );
    newlist.push_back( multiplyPointByMatrix( xform, ptlist[ 2 ] ) );

    Triangle *t = new Triangle( &newlist );
    t->setMaterial( getMaterial() );
    pl.push_back( t );
    return pl;
}

bool Triangle::closestIntersect( const Point3 &origin, const Vector3 &dir,
                                 double *t, Vector3 *norm, Material **mat ) const
{
    Vector3 pvec, tvec, qvec;
    double det, inv_det, u, v;

    /* begin calculating determinant - also used to calculate U parameter */
    pvec = dir.cross( edge2 );

    /* if determinant is near zero, ray lies in plane of triangle */
    det = edge1.dotProduct( pvec );

    if( det > -epsilon && det < epsilon ) return false;
    inv_det = 1.0 / det;

    /* calculate distance from vert0 to ray origin */
    tvec = origin.subtract( ptlist[ 0 ] );

    /* calculate U parameter and test bounds */
    u = tvec.dotProduct( pvec ) * inv_det;
    if( u < 0.0 || u > 1.0 ) return false;

    /* prepare to test V parameter */
    qvec = tvec.cross( edge1 );

    /* calculate V parameter and test bounds */
    v = dir.dotProduct( qvec ) * inv_det;
    if( v < 0.0 || u + v > 1.0 ) return false;

    /* calculate t, ray intersects triangle */
    *t = edge2.dotProduct( qvec ) * inv_det;

    if( *t < 0.0 ) return false;

    if( dir.dotProduct( normal ) < 0.0 ) {
        *norm = normal;
    } else {
        *norm = normal.scalarMult( -1.0 );
    }
    *mat = getMaterial();
    return true;
}


#if 0

// Original code

#define EPSILON 0.000001
#define CROSS(dest,v1,v2) \
          dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
          dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
          dest[2]=v1[0]*v2[1]-v1[1]*v2[0];
#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])
#define SUB(dest,v1,v2)
          dest[0]=v1[0]-v2[0]; \
          dest[1]=v1[1]-v2[1]; \
          dest[2]=v1[2]-v2[2]; 

int
intersect_triangle(double orig[3], double dir[3],
                   double vert0[3], double vert1[3], double vert2[3],
                   double *t, double *u, double *v)
{
   double edge1[3], edge2[3], tvec[3], pvec[3], qvec[3];
   double det,inv_det;

   /* find vectors for two edges sharing vert0 */
   SUB(edge1, vert1, vert0);
   SUB(edge2, vert2, vert0);

   /* begin calculating determinant - also used to calculate U parameter */
   CROSS(pvec, dir, edge2);

   /* if determinant is near zero, ray lies in plane of triangle */
   det = DOT(edge1, pvec);

#ifdef TEST_CULL           /* define TEST_CULL if culling is desired */
   if (det < EPSILON)
      return 0;

   /* calculate distance from vert0 to ray origin */
   SUB(tvec, orig, vert0);

   /* calculate U parameter and test bounds */
   *u = DOT(tvec, pvec);
   if (*u < 0.0 || *u > det)
      return 0;

   /* prepare to test V parameter */
   CROSS(qvec, tvec, edge1);

    /* calculate V parameter and test bounds */
   *v = DOT(dir, qvec);
   if (*v < 0.0 || *u + *v > det)
      return 0;

   /* calculate t, scale parameters, ray intersects triangle */
   *t = DOT(edge2, qvec);
   inv_det = 1.0 / det;
   *t *= inv_det;
   *u *= inv_det;
   *v *= inv_det;
#else                    /* the non-culling branch */
   if (det > -EPSILON && det < EPSILON)
     return 0;
   inv_det = 1.0 / det;

   /* calculate distance from vert0 to ray origin */
   SUB(tvec, orig, vert0);

   /* calculate U parameter and test bounds */
   *u = DOT(tvec, pvec) * inv_det;
   if (*u < 0.0 || *u > 1.0)
     return 0;

   /* prepare to test V parameter */
   CROSS(qvec, tvec, edge1);

   /* calculate V parameter and test bounds */
   *v = DOT(dir, qvec) * inv_det;
   if (*v < 0.0 || *u + *v > 1.0)
     return 0;

   /* calculate t, ray intersects triangle */
   *t = DOT(edge2, qvec) * inv_det;
#endif
   return 1;
}

#endif

