
#include <stdio.h>
#include <uniformgrid.h>

UniformGrid::UniformGrid( int gridwidth )
    : width( gridwidth * 4 ),
      minx( 0 ), maxx( 0 ), miny( 0 ), maxy( 0 ), minz( 0 ), maxz( 0 ),
      numprimitives( 0 ), populatedsquares( 0 ),
      xy( Vector3( 0.0, 0.0, 1.0 ) ),
      yz( Vector3( 1.0, 0.0, 0.0 ) ),
      zx( Vector3( 0.0, 1.0, 0.0 ) )
{
    squares = new list<GridSquare *>[ (width*2) * (width*2) ];
}

UniformGrid::~UniformGrid( void )
{
}

void UniformGrid::addPrimitive( Primitive *p )
{
    PrimitiveList::const_iterator it;
    BoundingBox b = p->getBoundingBox();
    int x, y, z;
    bool added = false;

    for( x = ((int) floor(b.minx))-1; x < ((int) floor(b.maxx))+1; x++ ) {
        for( y = ((int) floor(b.miny))-1; y < ((int) floor(b.maxy))+1; y++ ) {
            for( z = ((int) floor(b.minz))-1; z < ((int) floor(b.maxz))+1; z++ ) {
                GridSquare *sq = getSquare( x, y, z );
                bool found = false;

                for( it = sq->contents.begin(); it != sq->contents.end(); it++ ) {
                    if( (*it) == p ) { found = true; break; }
                }
                if( !found ) {
                    added = true;
                    sq->contents.push_back( p );
                }
            }
        }
    }
    if( added ) numprimitives++;
}

bool UniformGrid::gonFishin( const Point3 &pos ) const
{
    int x = (int) pos.getX();
    int y = (int) pos.getY();
    int z = (int) pos.getZ();

    return ( x < minx || x > maxx ||
             y < miny || y > maxy ||
             z < minz || z > maxz );
}

void UniformGrid::printBounds( void ) const
{
    list<GridSquare *>::const_iterator it;
    unsigned int largest = 0;
    unsigned int smallest = UINT_MAX;
    int i;

    for( i = 0; i < ( width * 2 * width * 2 ); i++ ) {
        list<GridSquare *> *curlist = &(squares[i]);
        for( it = curlist->begin(); it != curlist->end(); it++ ) {
            if( (*it)->contents.size() < smallest )
                smallest = (*it)->contents.size();
            if( (*it)->contents.size() > largest )
                largest = (*it)->contents.size();
        }
    }

    fprintf( stderr, "ugrid: Width %d, %d primitives, %d populated partitions.\n",
             width / 4, numprimitives, populatedsquares );
    fprintf( stderr, "ugrid: Largest partition contains %d primitives.\n", largest );
    fprintf( stderr, "ugrid: Smallest partition contains %d primitives.\n", smallest );
    fprintf( stderr, "ugrid: Extents: %d < x < %d, %d < y < %d, %d < z < %d.\n",
             minx, maxx, miny, maxy, minz, maxz );
}

GridSquare *UniformGrid::getCurrentSquare( const Point3 &pos ) const
{
    list<GridSquare *>::const_iterator it;
    int x = (int) floor(pos.getX());
    int y = (int) floor(pos.getY());
    int z = (int) floor(pos.getZ());
    list<GridSquare *> *curlist = &(squares[((y+width)*width*2) + (x+width)]);

    for( it = curlist->begin(); it != curlist->end(); it++ ) {
        if( (*it)->x == x && (*it)->y == y && (*it)->z == z ) {
            return (*it);
        }
    }

    return 0;
}

double UniformGrid::getNextSquare( const Point3 &pos,
                                   const Vector3 &dir ) const
{
    Point3 refpoint;
    int x = (int) floor(pos.getX());
    int y = (int) floor(pos.getY());
    int z = (int) floor(pos.getZ());
    double bestt = HUGE_VAL;
    double curt;

    refpoint = Point3( (double) x, (double) y, (double) z );

    if( dir.getZ() < 0.0 && xy.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    if( dir.getX() < 0.0 && yz.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    if( dir.getY() < 0.0 && zx.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    refpoint = Point3( (double) (x + 1), (double) (y + 1), (double) (z + 1) );

    if( dir.getZ() > 0.0 && xy.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    if( dir.getX() > 0.0 && yz.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    if( dir.getY() > 0.0 && zx.closestIntersect( refpoint, pos, dir, &curt )
                         && curt < bestt ) {
        bestt = curt;
    }

    return bestt + (2.0*epsilon);
}

GridSquare *UniformGrid::getSquare( int x, int y, int z )
{
    list<GridSquare *>::const_iterator it;
    list<GridSquare *> *curlist = &(squares[((y+width)*width*2) + (x+width)]);

    for( it = curlist->begin(); it != curlist->end(); it++ ) {
        if( (*it)->x == x && (*it)->y == y && (*it)->z == z ) {
            return (*it);
        }
    }

    // Not in the list.
    GridSquare *sq = new GridSquare;
    sq->x = x;
    sq->y = y;
    sq->z = z;
    if( (x-1) < minx ) minx = (x-1); if( (x+1) > maxx ) maxx = (x+1);
    if( (y-1) < miny ) miny = (y-1); if( (y+1) > maxy ) maxy = (y+1);
    if( (z-1) < minz ) minz = (z-1); if( (z+1) > maxz ) maxz = (z+1);
    curlist->push_back( sq );
    populatedsquares++;
    return sq;
}

